function data = FF_to_Matlab(datapath, d_inputs, outputs, cleanup)
% Convert the matrices generated by FreeFem++ in the directory given by 
% datapath into Matlab-readable format. The output is stored as a
% structure.
%
% Copyright 2020 Tony Stillfjord
%
% This file is part of BST20_CODE.
% 
% BST20_CODE is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
% 
% BST20_CODE is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
% 
% You should have received a copy of the GNU General Public License
% along with BST20_CODE.  If not, see <http://www.gnu.org/licenses/>.
%
if nargin < 2
   d_inputs = 3; 
end
if nargin < 3
   outputs = 2; 
end
if nargin < 4
    cleanup = false;
end

% Check what files are here, to see what should be converted
files = dir([datapath, '/K*.*']);
Nds = zeros(length(files),1);
for j = 1:length(files)
    Nds(j) = str2num(files(j).name(2:end-4));
end
Nds = sort(Nds)';

for Nd = Nds(1:end)
    Nd
    % Stiffness matrix
    K = FFmatrix_fread([datapath, 'K' , int2str(Nd) , '.txt']); 
    % Mass matrix
    M = FFmatrix_fread([datapath, 'M' , int2str(Nd) , '.txt']);
    % Boundary mass matrix
    MB = FFmatrix_fread([datapath, 'MB' , int2str(Nd) , '.txt']);
    
    N = size(K,1);

    [i,j] = find(K == 1e30);
    freeNodes = setdiff(1:N, i);
    Nfree = length(freeNodes);
    
    K = K(freeNodes, freeNodes); % Works in Matlab; Python needs K = K[freeNodes, :][:, freeNodes];
    M = M(freeNodes, freeNodes);
    MB = MB(freeNodes, freeNodes);
    
    K = -sparse(K);
    M = sparse(M);
    MB = sparse(MB);
    
    % Final state penalizer
    LT = FFvector_fread([datapath, 'LT' , int2str(Nd) , '.txt']);
    LT = LT(freeNodes);
    
	% Output matrix; C(j,k) = (\chi_j, \phi^h_k)_{L^2}
    Calt = [];
    for j = 1:outputs
        chi{j} = FFvector_fread([datapath, 'chi', int2str(j), '_' , int2str(Nd) , '.txt']);
        chi{j} = chi{j}(freeNodes);
        
        Calt = [Calt, chi{j}];
    end
    % The above are the FEM-coordinate vectors of the L^2-orthogonal projections.
    % To get (chij, phi^h_k), multiply with the mass matrix:
    Calt = (M*Calt)';

    C = FFmatrix_fread([datapath, 'C' , int2str(Nd) , '.txt']);
    C = C';
    
    % Input matrices
    % Distributed controls; B(k,j) = (P_h \psi_j, \phi^h_k)_{L^2} = (M P_h \psi_j)_k
    BDalt = [];
    for j = 1:d_inputs
        psi{j} = FFvector_fread([datapath, 'psi', int2str(j), '_' , int2str(Nd) , '.txt']);
        psi{j} = psi{j}(freeNodes);
        BDalt = [BDalt, psi{j}];
    end
    % The above are the FEM-coordinate vectors of the L^2-orthogonal projections.
    % To get (psij, phi^h_k), multiply with the mass matrix:
    BDalt = M*BDalt;
    
    BD = FFmatrix_fread([datapath, 'BD' , int2str(Nd) , '.txt']);
    
    % Boundary controls; B(k,j) = ((lambda*I - A) R_h \Nv, \phi^h_k)_{L^2}
    
    % RhNv = \sum_{j=1}^{4}{RhNv_j}, where the four RhNv_j come from the
    % four different sides of the unit square
    RhNv1 = full(FFmatrix_fread([datapath, 'Rh1' , int2str(Nd) , '.txt']));
    RhNv2 = full(FFmatrix_fread([datapath, 'Rh2' , int2str(Nd) , '.txt']));
    RhNv3 = full(FFmatrix_fread([datapath, 'Rh3' , int2str(Nd) , '.txt']));
    RhNv4 = full(FFmatrix_fread([datapath, 'Rh4' , int2str(Nd) , '.txt']));

    lambda = 1;

    RhNv = RhNv1(:,1) + RhNv2(:,1) + RhNv3(:,1) + RhNv4(:,1);
    % By multiplying with lambda*M + K (corresponding to \lambda*I -\Delta),
    % we get do not get the FEM coordinate vectors of \lambda RhNv -\Delta RhNv 
    % but rather (\lambda RhNv -\Delta RhNv, \phi_k), k = 1, ..., N_h.
    % However, this is exactly what we need.
    BN = (lambda*M - K)*RhNv(:,1);
    
    % Alternatively, control each side separately
    RhNv_sep = [RhNv1(:,1), RhNv2(:,1), RhNv3(:,1), RhNv4(:,1)];
    BN4 = (lambda*M - K)*RhNv_sep;
       
    B = [BD, BN];
   
    % Compute some L^2-norms
    for j = 1:size(BD, 2)
        data.BDnorm(j) = L2norm(BD(:,j), M);
    end
    for j = 1:size(BN4, 2)
        data.BN4norm(j) = L2norm(BN4(:,j), M);
    end
    data.BNnorm = L2norm(BN, M);
        

    % Interpolation matrix
    Ih = FFmatrix_fread([datapath, 'Ih' , int2str(Nd) , '.txt']);
    
    % Projection matrix (between discretizations, not from L^2(G))
    Ph = FFmatrix_fread([datapath, 'Ph' , int2str(Nd) , '.txt']);
    
    % Mesh
    [p, t, e, ~] = loadmesh([datapath, 'mesh' , int2str(Nd) , '.msh']);
    
    
    data.K = K;
    data.M = M;
    data.MB = MB;

    data.BD = BD;
    data.BDalt = BDalt;
    data.BN = BN;
    data.BN4 = BN4;
    data.B = B;
    
    data.C = C;
    data.Calt = Calt;
    
    data.LT = LT;
    
    data.chi = chi;
    data.psi = psi;
    
    data.RhNv1 = RhNv1;
    data.RhNv2 = RhNv2;
    data.RhNv3 = RhNv3;
    data.RhNv4 = RhNv4;
    data.RhNv = RhNv;
    
    data.Ih = Ih;
    data.Ph = Ph;
    
    data.N = N;
    data.Nfree = Nfree;
    data.mesh.p = p;
    data.mesh.e = e;
    data.mesh.t = t;
    
    save([datapath, 'data', int2str(Nd), '.mat'], '-struct', 'data')
    
end

if cleanup
    % Clear up temporary files
    delete([datapath, '*.txt'])
    delete([datapath, '*.msh'])
end