function Zj = IE_LQR(M, K, G, dW, Z0, Nt, dt, FEMdata, B, Ks)
% Solve the system
%   M dx = Kx dt + G(dW)
% if B and Ks are not given. If they are, solve instead
%   M dx = (Kx - B*Ks*x) dt + G(dW)
% Here Ks are the feedback matrices. Ks{j} should contain the feedback
% matrix at time t_j = t0 + j*dt.
%
% If dW is a matrix then this is used for distributed noise only.
% If dW is a cell-array with two elements, the first element is used as the
% distributed noise and the second element is used as boundary noise. The
% modified equation will be
%   M dx = (Kx - B*Ks*x) dt + G{1}(dW{1}) + G{2}(dW{2})
%
% It is assumed that e.g. (lambda*I - A) has already been applied to the
% noise.  G{1} and G{2} are transformation functions that currently only
% scale the noise with a (small) constant factor.
%
% Copyright 2020 Tony Stillfjord
%
% This file is part of BST20_CODE.
% 
% BST20_CODE is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
% 
% BST20_CODE is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
% 
% You should have received a copy of the GNU General Public License
% along with BST20_CODE.  If not, see <http://www.gnu.org/licenses/>.
%
if nargin < 9
    control = false;
else
    control = true;
end

if iscell(dW) 
    if length(dW) > 1
        boundary_noise = true;
    else
        boundary_noise = false;
        dW = dW{1};
    end
else
    boundary_noise = false;
end

Zj = zeros(size(Z0, 1), Nt+1, size(Z0, 2));
Zj(:,1,:) = Z0;


J = M - dt*K;
if control
    SMW = J\B;
    I = eye(size(Ks{1}, 1));
end

for j = 2:Nt+1
    if ~boundary_noise
        RHS = M*squeeze(Zj(:,j-1,:)) + M*squeeze(G(dW(:,j-1,:)));
    else
        RHS = M*squeeze(Zj(:,j-1,:)) + M*squeeze(G{1}(dW{1}(:,j-1,:))) + M*squeeze(G{2}(dW{2}(:,j-1,:)));    
    end
    
    if control
%         J = M - dt*(K - B*Ks{j});
%         Zj(:,j,:) = J\RHS;
        JinvRHS = J\RHS;
        Zj(:,j,:) = JinvRHS - dt*SMW*((I + dt*Ks{j}*SMW)\(Ks{j}*JinvRHS));
        
    else
        Zj(:,j,:) = J\RHS;
    end

end

end