function dQW = QWiener1D_FEM_implicit(Nt, dt, Nx, a, N, ind, MB, params)
% Use DCT to generate a FEM discretization of 1D noise on [0, 4*a] then 
% wrap it around the boundary of the square [0, a]^2 by extending the FEM
% function to its 2D representation and applying the boundary mass matrix.
%
% ind : the 2D indices of the boundary vertices
% MB : the boundary mass matrix
%
% Copyright 2020 Tony Stillfjord
%
% This file is part of BST20_CODE.
% 
% BST20_CODE is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
% 
% BST20_CODE is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
% 
% You should have received a copy of the GNU General Public License
% along with BST20_CODE.  If not, see <http://www.gnu.org/licenses/>.
%

Nxb = 4*Nx - 4; % Number of boundary vertices

% Eigenvalues 1D
if isfield(params, 'l')
    l = params.l;
    mu_j = 2 ./ l ./ (pi^2*(1:Nxb)'/a + 1/l^2);

elseif isfield(params, 'beta')
    beta = params.beta;
    eps = params.eps;
    mu_j = ((1:Nxb)'.^2).^ (-(beta + eps));
   
elseif isfield(params, 'r')
    r = params.r;
    eps = params.eps;
    mu_j = (1:Nxb)'.^(-(2*r+1+eps));
end


% Gaussian noise
dW = randn(Nxb, Nt, N);

% Expansion coefficients
bj = sqrt(mu_j)*sqrt(dt)*Nxb;

% Form bj*dW
tmphat = bsxfun(@times, bj, dW);
% % For testing:
% tmphat = zeros(size(tmphat)); 
% tmphat(5,:,:) = 1;

% Do a discrete cosine transform (DCT-1) in the spatial dimension
dQW_1D = dct(tmphat, [], 1, 'Type', 1);
dQW_1D = dQW_1D*sqrt(Nxb-1)/sqrt(2);

% Compensate for peculiar scaling:
dQW_1D(1, :, :) = dQW_1D(1, :, :) * sqrt(2);
dQW_1D(end, :, :) = dQW_1D(end, :, :) * sqrt(2);

dQW = zeros(Nx^2, Nt, N);
dQW(ind, :, :) = dQW_1D;

for j = 1:Nt
    dQW(:, j, :) = MB * squeeze(dQW(:,j,:));
end

end