function [dQW, dQW_coarse] = QWiener2D_FEM_DCT(Nt, dt, Nx, Ny, a1, a2, N, params, projection)
% Form 2D distributed noise via a two-fold DCT as in Section 10.2 of 
% Lord, Powell and Shardlow, "An Introduction to Computational Stochastic
% PDEs", Cambridge University Press, 2014, DOI 10.1017/CBO9781139017329.
%
% If the projection argument is supplied, it is assumed that the number of 
% time steps increase by a factor of 4 at each level and the number of 
% spatial grid points increase by a factor 2. Under this assumption, also 
% the noise projected onto both the coarser temporal grid and the coarser 
% spatial grid is returned. The projection argument is supposed to be the 
% transformation matrix performing the V_l -> V_{l-1} operation.
%
% Copyright 2020 Tony Stillfjord
%
% This file is part of BST20_CODE.
% 
% BST20_CODE is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
% 
% BST20_CODE is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
% 
% You should have received a copy of the GNU General Public License
% along with BST20_CODE.  If not, see <http://www.gnu.org/licenses/>.
%

% Eigenvalues 2D
if isfield(params, 'l')
    l = params.l;
    mu_j_x = 2 ./ l ./ (pi^2*(1:Nx)'/a1 + 1/l^2);
    mu_j_y = 2 ./ l ./ (pi^2*(1:Ny)'/a2 + 1/l^2);
    mu_j = mu_j_x*mu_j_y';

elseif isfield(params, 'beta')
    beta = params.beta;
    eps = params.eps;
    [I, K] = meshgrid(1:Nx, 1:Ny);
    
    mu_j = (I.^2 + K.^2).^ (-(beta + eps)) / a1 / a2;
   
elseif isfield(params, 'r')
    r = params.r;
    eps = params.eps;
    [I, K] = meshgrid(1:Nx, 1:Ny);
    
    mu_j = (I.*K).^(-(2*r+1+eps)) / a1 / a2;
end

% Gaussian noise
dW = randn(Nx, Ny, Nt, N);

% Computing the bj coefficients
% bj = sqrt(mu_j)*sqrt(dt)*Nx*Ny/sqrt(a1*a2);
bj = sqrt(mu_j)*sqrt(dt)/sqrt(a1*a2);

% Form bj*dW
tmphat = bsxfun(@times, bj, dW);

% Do a discrete cosine transform (DCT-1) in each spatial dimension
dQW = dct(dct(tmphat, [], 2, 'Type', 1), [], 1, 'Type', 1);
dQW = dQW*sqrt((Nx-1)*(Ny-1))/2;

% Compensate for peculiar scaling:
dQW(1, :, :, :) = dQW(1, :, :, :) * sqrt(2);
dQW(end, :, :, :) = dQW(end, :, :, :) * sqrt(2);
dQW(:, 1, :, :) = dQW(:, 1, :, :) * sqrt(2);
dQW(:, end, :, :) = dQW(:, end, :, :) * sqrt(2);

dQW = reshape(dQW, Nx*Ny, Nt, N);

if nargin > 8
    % Temporal projection
    dQW_coarse_temp = dQW(:,1:4:end-3, :) + dQW(:,2:4:end-2, :) + dQW(:,3:4:end-1, :) + dQW(:,4:4:end, :);

    % Project noise onto V(l-1):
    dQW_coarse = zeros(Nx*Ny/4, Nt/4, N);
    for j = 1:N
        dQW_coarse(:,:,j) = projection*dQW_coarse_temp(:,:,j);
    end

end

end