function out = compute_optimal_feedback(data, FEMdata, save_solution)
% Copyright 2020 Tony Stillfjord
%
% This file is part of BST20_CODE.
% 
% BST20_CODE is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
% 
% BST20_CODE is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
% 
% You should have received a copy of the GNU General Public License
% along with BST20_CODE.  If not, see <http://www.gnu.org/licenses/>.
%

if nargin < 3
    save_solution = false;
end

for L = data.Ls
    Nx = data.Nx(L);
    
    if ~iscell(FEMdata) % only one spatial discretization
        problem = FEMdata;
    else
        problem = FEMdata{L};
    end
        
    problem.A = data.diff_coeff*problem.K; % DREsplit notation
    
    problem.B = full(problem.B);
    problem.C = full(problem.C);

    % LDLT-factorization of Q = C'C
    problem.LQ = problem.C';
    problem.DQ = data.Q * eye(size(problem.C', 2));

    % LDLT-factorization of initial condition P(0) = 0
%     L0 = zeros(Nx^2, 1);
%     L0 = problem.C'

    L0 = problem.LT;
    D0 = data.RT * eye(size(L0, 2));

    problem.L0 = L0;
    problem.D0 = D0;

    problem.Rinv = data.R \ eye(size(problem.B, 2));

    % Column compression tolerance
    problem.CC_tol = Nx*eps;

    % Integration interval
    problem.t0 = 0;
    problem.tend = data.T;
    problem.Nt = data.Nt(L);

    % Options for computing e^{hA}x terms
%     problem.impODE.method = 'exact';    % Use expm 
    problem.impODE.method = 'krylov';    % Use polynomial Krylov
    problem.impODE.tol = 1e-6;
%     problem.impODE.method = 'IRKadaptive';
%     problem.impODE.tol = 1e-6;

%     problem.quadrature.type = 'gauss'; 
%     problem.quadrature.order = 7; % optional; the methods choose if not supplied
    problem.quadrature.type = 'adaptive'; 
    problem.quadrature.tol = 1e-3; % tolerance for above adaptivity
%    Absolute rather than relative error in quadrature computation, and the 
%    norm of the integral is pretty big, so this means basically round-off
%    accurate:
%     problem.quadrature.tol = 1e-9; % tolerance for above adaptivity 


    % Output options 
    problem.intermediates = true;      % Store intermediate steps 
    problem.print_step = true;         % Print number of total steps and steps taken
    problem.print_output = false;      % Print extra info every step

    problem.save_solution = save_solution;
    problem.solver = 'Strang';
    
    problem
    sol = general_solver(problem);

    Kss{L} = sol.Ks;
    if save_solution
        Lss{L} = sol.Ls;
        Dss{L} = sol.Ds;
    end
end

out.Kss = Kss;
out.Rinv = problem.Rinv;
if save_solution
    out.Lss = Lss;
    out.Dss = Dss;
end

save([data.outputpath, 'DREsol_', data.name, '_Nx_', int2str(data.Nx(end)), '_Nt_', int2str(data.Nt(end)), '.mat'], '-struct', 'out', '-v7.3')

end