function main_test_spacetime_with_parameters(Lmax, N, Nbatch, reuse_DREsol)
% Copyright 2020 Tony Stillfjord
%
% This file is part of BST20_CODE.
% 
% BST20_CODE is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
% 
% BST20_CODE is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
% 
% You should have received a copy of the GNU General Public License
% along with BST20_CODE.  If not, see <http://www.gnu.org/licenses/>.
%

if nargin < 4
    reuse_DREsol = false;
end

addpath('../DREsplit')

data = setup_problem_spacetime(Lmax, true);

Lmax
N
Nbatch
data.name
reuse_DREsol

%% Load matrices
for j = 1:Lmax
    FEMdata{j} = load([data.path, 'data', int2str(data.Nx(j)), '.mat']);
    % Override input operator to ignore the boundary input:
    % FEMdata{j}.B = FEMdata{j}.BD;
    % data.R = data.Rd;
end


%% Compute optimal feedback by solving DRE
if ~reuse_DREsol
    DREdata = compute_optimal_feedback(data, FEMdata);
else
    DREdata = load([data.outputpath, 'DREsol_', data.name, '_Nx_', int2str(data.Nx(end)), '_Nt_', int2str(data.Nt(end)), '.mat']);
end

%%
rng('default') % Ensure same noise is used

Nbatches = ceil(N / Nbatch);
N = Nbatches*Nbatch;
for j = 1:Lmax
    pMS{j} = zeros(data.Nt(j) + 1, 1);
    pMSc{j} = zeros(data.Nt(j) + 1, 1);
end

cost_ref = 0;
cost_cref = 0;

for i = 1:Nbatches
    fprintf('Batch %d of %d\n', i, Nbatches)
    % Integrate both uncontrolled and controlled systems Nbatch times for
    % different step sizes
    z0 = L2projection_continuous(data.z0, FEMdata{Lmax});
    Z0 = repmat(z0, 1, Nbatch); % Initial conditions for all the sampled systems

    errs = zeros(Lmax-1, 1);
    errsc = zeros(Lmax-1, 1);

    % Generate fine-scale distributed noise paths
    dQW = QWiener2D_FEM_DCT(data.Nt(Lmax), data.dt(Lmax), data.Nx(Lmax)+1, data.Nx(Lmax)+1, data.a, data.a, Nbatch, data.noise_params_D);
    fprintf('Generated distributed noise\n')

    % Generate fine-scale boundary noise paths
    dQW_bdy = QWiener1D_FEM_explicit(data.Nt(Lmax), data.dt(Lmax), data.Nx(Lmax)+1, data.a, data.X{Lmax}, data.Y{Lmax}, data.lambda, ...
                                     FEMdata{Lmax}.M, FEMdata{Lmax}.K, Nbatch, data.noise_params_N, true, FEMdata{Lmax});
    fprintf('Generated boundary noise\n')

    dW{1} = dQW;
    dW{2} = dQW_bdy;
   
    G{1} = data.GD;
    G{2} = data.GN;
    
    % Feedback matrices for controlled system
    % (DREsplit changes the direction of time so the solution has to be
    % evaluated backwards)
    Ks = fliplr(DREdata.Kss{Lmax});

    % Uncontrolled reference solution
    tic
    Zref = IE_LQR(FEMdata{Lmax}.M, data.diff_coeff*FEMdata{Lmax}.K, G, dW, Z0, data.Nt(Lmax), data.dt(Lmax), FEMdata{Lmax});
    toc
    fprintf('Computed uncontrolled reference solution\n')

    % Controlled reference solution
    tic
    Zcref = IE_LQR(FEMdata{Lmax}.M, data.diff_coeff*FEMdata{Lmax}.K, G, dW, Z0, data.Nt(Lmax), data.dt(Lmax), FEMdata{Lmax}, FEMdata{Lmax}.B, Ks);
    toc
    fprintf('Computed controlled reference solution\n')

    for j = Lmax-1:-1:1 % loop over discretizations
        j
        % Find noise on coarser temporal grid
        dW{1} = dW{1}(:, 1:2:end-1, :) + dW{1}(:, 2:2:end, :);
        dW{2} = dW{2}(:, 1:2:end-1, :) + dW{2}(:, 2:2:end, :);
		% Repeat, because Nt = 2^(2*L), L = 1, 2, ..., Lmax
		dW{1} = dW{1}(:, 1:2:end-1, :) + dW{1}(:, 2:2:end, :);
        dW{2} = dW{2}(:, 1:2:end-1, :) + dW{2}(:, 2:2:end, :);
		
		% Project noise onto next coarser spatial grid:
        dW{1} = L2projection_FEM_batched(dW{1}, FEMdata{j+1}.Ih, FEMdata{j}.M, FEMdata{j+1}.M);
        dW{2} = L2projection_FEM_batched(dW{2}, FEMdata{j+1}.Ih, FEMdata{j}.M, FEMdata{j+1}.M);

		z0 = L2projection_continuous(data.z0, FEMdata{j});
        Z0 = repmat(z0, 1, Nbatch); % Initial conditions for all the sampled systems
		
        Z = IE_LQR(FEMdata{j}.M, data.diff_coeff*FEMdata{j}.K, G, dW, Z0, data.Nt(j), data.dt(j), FEMdata{j});

        Ks = fliplr(DREdata.Kss{j});
        Zc = IE_LQR(FEMdata{j}.M, data.diff_coeff*FEMdata{j}.K, G, dW, Z0, data.Nt(j), data.dt(j), FEMdata{j}, FEMdata{j}.B, Ks);

        Zref_coarse_time = Zref(:,1:data.Nt(Lmax)/data.Nt(j):end,:);
        Zcref_coarse_time = Zcref(:,1:data.Nt(Lmax)/data.Nt(j):end,:);

		% Interpolate current approximation to finest spatial grid in order
        % to compute difference to reference approximation in the L^2-norm
        Z_fine_space = Z;
        Zc_fine_space = Zc;
        for k = j:Lmax-1
            Z_fine_space = lift_batched(FEMdata{k+1}.Ih, Z_fine_space);
            Zc_fine_space = lift_batched(FEMdata{k+1}.Ih, Zc_fine_space);
        end
		
        % We compute all the (squared) pointwise errors first then compute the
        % respective temporal norm afterwards. Otherwise, we cannot update the
        % pointwise errors with the new batch of samples.
        pMS{j} = pMS{j} + Nbatch * pointwise_RMS_L2_norm(Z_fine_space - Zref_coarse_time, FEMdata{Lmax}.M) .^2;
        pMSc{j} = pMSc{j} + Nbatch * pointwise_RMS_L2_norm(Zc_fine_space - Zcref_coarse_time, FEMdata{Lmax}.M) .^2;
    end
    
    % Compute cost-functionals
    for j = 1:Nbatch
        cost_ref_j = cost_functional(data.Nt(end), data.dt(end), FEMdata{Lmax}.C, Zref(:,:,j), data.Q, Lmax, data.RT);
        cost_ref = cost_ref + cost_ref_j;
        cost_cref_j = cost_functional(data.Nt(end), data.dt(end), FEMdata{Lmax}.C, Zcref(:,:,j), data.Q, Lmax, data.RT, data.R, fliplr(DREdata.Kss{end}));
        cost_cref = cost_cref + cost_cref_j;
    end
    
end

%% Finalize pointwise error computation
for j = Lmax-1:-1:1
    pMS{j} = sqrt(pMS{j} / N);
    pMSc{j} = sqrt(pMSc{j} / N);
end
%%

%%
% Compute errors. All versions use the RMS-Lmax^2 norm in space, i.e.
% if Z = [Z_1, ..., Z_N] are the different samples then
% ||Z|| = sqrt( 1/N \sum_{j=1}^{N} { ||Z_j||_{L^2}^2 }.
% In time, we use either max-norm;
%   max_{k=1,...,Nt} m||Z(t_k)||,
% L^2-norm;
%   sqrt( dt \sum_{i=1}^{Nt} { ||Z(t_i)||^2 } ),
% or point-evaluation;
%   ||Z(t)||.
for j = Lmax-1:-1:1
    % max-norm
    [errs_max(j), maxind(j)] = max(pMS{j});
    [errsc_max(j), maxindc(j)] = max(pMSc{j});

    % L2-norm
    errs_L2(j) = sqrt( data.dt(j) * sum(pMS{j}.^2));
    errsc_L2(j) = sqrt( data.dt(j) * sum(pMSc{j}.^2));

    % Pointwise at T/4
    errs_Td4(j) = pMS{j}(data.Nt(j)/4 + 1);
    errsc_Td4(j) = pMSc{j}(data.Nt(j)/4 + 1);

    % Pointwise at T/2
    errs_Td2(j) = pMS{j}(data.Nt(j)/2 + 1);
    errsc_Td2(j) = pMSc{j}(data.Nt(j)/2 + 1);
    
    % Pointwise at T
    errs_T(j) = pMS{j}(data.Nt(j) + 1);
    errsc_T(j) = pMSc{j}(data.Nt(j) + 1);
end

%% Compute a trajectory
plotsample = 1;

u = zeros(size(FEMdata{Lmax}.B, 2), data.Nt(end));
Ks = fliplr(DREdata.Kss{end});
for j = 1:data.Nt(end)
    u(:,j) = Ks{j}*Zcref(:,j,plotsample);
end

yc = FEMdata{Lmax}.C*Zcref(:,:,plotsample);
y = FEMdata{Lmax}.C*Zref(:,:,plotsample);


samplecost_c = cost_functional(data.Nt(end), data.dt(end), FEMdata{Lmax}.C, Zcref(:,:,plotsample), data.Q, Lmax, data.RT, data.R, fliplr(DREdata.Kss{end}))
samplecost = cost_functional(data.Nt(end), data.dt(end), FEMdata{Lmax}.C, Zref(:,:,plotsample), data.Q, Lmax, data.RT)

cost_cref
cost_ref
cost_cref / cost_ref

results.samplecost = samplecost;
results.samplecost_c = samplecost_c;

results.cost_ref = cost_ref;
results.cost_cref = cost_cref;

results.u = u;

results.y = y;
results.yc = yc;

results.errs_max = errs_max;
results.errsc_max = errsc_max;
results.errs_L2 = errs_L2;
results.errsc_L2 = errsc_L2;
results.errs_T = errs_T;
results.errsc_T = errsc_T;
results.errs_Td2 = errs_Td2;
results.errsc_Td2 = errsc_Td2;
results.errs_Td4 = errs_Td4;
results.errsc_Td4 = errsc_Td4;

results.pMS = pMS;
results.pMSc = pMSc;

save([data.outputpath, 'results_', int2str(data.Nt(end)), '_', int2str(data.Nx(end)), '.mat'], '-struct', 'results', '-v7.3')

% Zref_sample = Zref(:, :, plotsample);
% Zcref_sample = Zcref(:, :, plotsample);
% save([data.outputpath, 'Zref_sample_', int2str(data.Nt(end)), '_', int2str(data.Nx(end)), '.mat'], 'Zref_sample', '-v7.3')
% save([data.outputpath, 'Zcref_sample_', int2str(data.Nt(end)), '_', int2str(data.Nx(end)), '.mat'], 'Zcref_sample', '-v7.3')

end